using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.SceneManagement;

enum State {
    
    none,
    
    intro,
    tetromino,
    game,
    
    end,
    
    count
};

enum Stage {
    
    none,
    
    init,
    running,
    
    count
};

class Timer {
    public float current;
    public float duration;
};

public class main : MonoBehaviour {
    
    public GameObject player;
    public Transform cam;
    public Transform camera_target;
    public GameObject[ ] tetrominos;
    public GameObject board;
    public GameObject arche;
    public Material preview_ok;
    public Material preview_ko;
    public Material tetrominos_trans;
    public GameObject board_cube;
    public GameObject[ ] lifes;
    public GameObject text;
    
    GameObject[ ] arches;
    Vector3 camera_target_offset;
    Vector3 camera_ship_offset;
    float speed = 100.0f;
    GameObject[ ] pieces;
    int[ ] piece_indices;
    GameObject[ ] previews;
    GameObject[ ] board_cubes;
    
    int ship_x = 0;
    int ship_y = 0;
    float arches_spacing = 50.0f;
    
    State state = State.none;
    Stage stage = Stage.none;
    Timer wait;
    
    int[] board_values;
    
    GameObject ship_piece;
    int ship_piece_index;
    int ship_piece_orientation;
    int life;
    bool can_hit;
    float game_duration;
    
    void Start( ) {
        
        pieces = new GameObject[ 3 ];
        piece_indices = new int[ 3 ];
        previews = new GameObject[ tetrominos.Length ];
        
        for ( int i = 0; i < previews.Length; i++ ) {
            previews[ i ] = Instantiate( tetrominos[ i ], new Vector3( 0, 0, 0 ), Quaternion.identity );
            previews[ i ].SetActive( false );
        }
        
        arches = new GameObject[ 12 ];
        float zStart = 300.0f;
        
        for ( int i = 0; i < arches.Length; i++ ) {
            arches[ i ] = Instantiate( arche, new Vector3( 0, 0, zStart + i * arches_spacing ), Quaternion.identity );
        }
        
        board_cubes = new GameObject[ 100 ];
        for ( int i = 0, cube_y = -5; cube_y < 5; cube_y++ ) {
            
            for ( int cube_x = -5; cube_x < 5; cube_x++ ) {
                board_cubes[ i ] = Instantiate( board_cube, new Vector3( 0, 0, 0 ), Quaternion.identity );
                board_cubes[ i ].transform.parent = board.transform;
                board_cubes[ i ].transform.position = new Vector3( cube_x * 4 + 2, cube_y * 4 + 2 );
                board_cubes[ i ].SetActive( false );
                i++;
            }
        }
        
        board_values = new int[ 100 ];
        
		camera_target_offset = camera_target.position - cam.position;
        camera_ship_offset = cam.position - player.transform.position;
        player.transform.position = new Vector3( ship_x * 4 + 2, ship_y * 4 + 2 );
        cam.position = player.transform.position + camera_ship_offset;
        camera_target.position = cam.position + camera_target_offset;
        
        life = 3;
        can_hit = true;
        
        wait = new Timer( );
        state_switch( State.intro );
        game_duration = 0;
    }
	
    void Update( ) {
        
        switch ( state ) {
            
            case State.intro: {
                
                game_duration += Time.deltaTime;
                
                if ( state_in_init_stage( ) ) {
                    timer_start( wait, 5 );
                }
                
                ship_update( );
                arch_update( );
                timer_update( wait );
                
                if ( timer_isFinished( wait, true ) ) {
                    state_switch( State.tetromino );
                }
                
            } break;
            
            case State.tetromino: {
                
                game_duration += Time.deltaTime;
                
                if ( state_in_init_stage( ) ) {
                    tetromino_spawn( new Vector3( 0, 0, player.transform.position.z + arches_spacing * arches.Length ) );
                }
                
                ship_update( );
                arch_update( );
                tetromino_update( );
                
                if ( ship_piece ) {
                    state_switch( State.game );
                } else {
                    if ( pieces[ 0 ].transform.position.z < cam.position.z ) {
                        tetromino_spawn( new Vector3( 0, 0, player.transform.position.z + arches_spacing * arches.Length ) );
                    }
                }
                
            } break;
            
            case State.game: {
                
                game_duration += Time.deltaTime;
                
                if ( state_in_init_stage( ) ) {
                    board.transform.position = new Vector3( 0, 0, player.transform.position.z + arches_spacing * arches.Length );
                    board.SetActive( true );
                    tetromino_spawn( board.transform.position + new Vector3( 0, 0, arches_spacing * arches.Length * 0.5f ) );
                }
                
                ship_update( );
                arch_update( );
                tetromino_update( );
                
                Vector3 board_position = board.transform.position;
                board_position.z -= speed * Time.deltaTime;
                
                if ( board_position.z < cam.position.z ) {
                    board_position = new Vector3( 0, 0, player.transform.position.z + arches_spacing * arches.Length );
                    can_hit = true;
                }
                
                board.transform.position = board_position;
                Vector3 position = player.transform.position + new Vector3( 0, 0, 15 );
                bool add = Mathf.Abs(  position.z - board_position.z ) < speed * Time.deltaTime;
                
                if ( ship_piece ) {
                    
                    ship_piece.transform.position = position;
                    ship_piece.GetComponentInChildren<Renderer>( ).material = tetrominos_trans;
                    
                    GameObject preview = previews[ ship_piece_index ];
                    preview.SetActive( true );
                    
                    switch ( ship_piece_orientation ) {
                        
                        case 0: {
                            ship_piece.transform.rotation = Quaternion.AngleAxis( 0, Vector3.forward );
                            preview.transform.rotation = Quaternion.AngleAxis( 0, Vector3.forward );
                        } break;
                        
                        case 1: {
                            ship_piece.transform.rotation = Quaternion.AngleAxis( 90, Vector3.forward );
                            preview.transform.rotation = Quaternion.AngleAxis( 90, Vector3.forward );
                        } break;
                        
                        case 2: {
                            ship_piece.transform.rotation = Quaternion.AngleAxis( 180, Vector3.forward );
                            preview.transform.rotation = Quaternion.AngleAxis( 180, Vector3.forward );
                        } break;
                        
                        case 3: {
                            ship_piece.transform.rotation = Quaternion.AngleAxis( 270, Vector3.forward );
                            preview.transform.rotation = Quaternion.AngleAxis( 270, Vector3.forward );
                        } break;
                    }
                    
                    Piece piece = new Piece( ship_piece_index, ship_piece_orientation );
                    bool piece_fit = does_piece_fit_in_board( piece, board_values, ship_x, ship_y );
                    
                    if ( piece_fit ) {
                        
                        preview.GetComponentInChildren<Renderer>( ).material = preview_ok;
                        
                        if ( add ) {
                            add_piece_in_board( piece, board_values, ship_x, ship_y );
                            preview.SetActive( false );
                            Destroy( ship_piece );
                            ship_piece = null;
                            ship_piece_orientation = 0;
                            can_hit = false;
                        }
                        
                    } else {
                        
                        preview.GetComponentInChildren<Renderer>( ).material = preview_ko;
                        
                        if ( add ) {
                            life--;
                            speed -= 5;
                            preview.SetActive( false );
                            Destroy( ship_piece );
                            ship_piece = null;
                            ship_piece_orientation = 0;
                            can_hit = false;
                        }
                    }
                    
                    position.z = board.transform.position.z - 2;
                    preview.transform.position = position;
                }
                
                if ( add && can_hit ) {
                    life--;
                    speed -= 5;
                    can_hit = false;
                }
                
                for ( int i = 0; i < pieces.Length; i++ ) {
                    
                    if ( pieces[ i ] && pieces[ i ].transform.position.z < cam.position.z ) {
                        tetromino_spawn( new Vector3( 0, 0, board_position.z + arches_spacing * arches.Length * 0.5f ) );
                        break;
                    }
                }
                
                int[] to_remove = new int[ 100 ];
                bool speed_increase = false;
                
                for ( int y = 0; y < 10; y++ ) {
                    
                    bool ok = true;
                    for ( int x = 0; ok && x < 10; x++ ) {
                        if ( board_values[ y * 10 + x ] == 0 ) {
                            ok = false;
                        }
                    }
                    
                    if ( ok ) {
                        speed_increase = true;
                        for ( int x = 0; x < 10; x++ ) {
                            to_remove[ y * 10 + x ] = 1;
                        }
                    }
                }
                
                for ( int x = 0; x < 10; x++ ) {
                    
                    bool ok = true;
                    for ( int y = 0; ok && y < 10; y++ ) {
                        if ( board_values[ y * 10 + x ] == 0 ) {
                            ok = false;
                        }
                    }
                    
                    if ( ok ) {
                        speed_increase = true;
                        for ( int y = 0; y < 10; y++ ) {
                            to_remove[ y * 10 + x ] = 1;
                        }
                    }
                }
                
                if ( speed_increase ) {
                    speed += 5;
                }
                
                for ( int i = 0; i < board_values.Length; i++ ) {
                    
                    if ( to_remove[ i ] == 1 ) {
                        board_values[ i ] = 0;
                    }
                    
                    if ( board_values[ i ] != 0 ) {
                        board_cubes[ i ].SetActive( true );
                    } else {
                        board_cubes[ i ].SetActive( false );
                    }
                }
                
                switch ( life ) {
                    case 0: {
                        lifes[ 0 ].SetActive( false );
                        lifes[ 1 ].SetActive( false );
                        lifes[ 2 ].SetActive( false );
                        state_switch( State.end );
                    } break;
                    case 1: {
                        lifes[ 0 ].SetActive( true );
                        lifes[ 1 ].SetActive( false );
                        lifes[ 2 ].SetActive( false );
                    } break;
                    case 2: {
                        lifes[ 0 ].SetActive( true );
                        lifes[ 1 ].SetActive( true );
                        lifes[ 2 ].SetActive( false );
                    } break;
                    case 3: {
                        lifes[ 0 ].SetActive( true );
                        lifes[ 1 ].SetActive( true );
                        lifes[ 2 ].SetActive( true );
                    } break;
                }
                
            } break;
            
            case State.end: {
                
                text.transform.rotation = cam.transform.rotation;
                text.transform.position = cam.transform.position + new Vector3( 0, 0, 10 );
                text.SetActive( true );
                text.GetComponent<TextMesh>( ).text = "Your time:\n" + game_duration + "s\n\nPress R to restart.";
                
                if ( Input.GetKeyDown( KeyCode.R ) ) {
                    SceneManager.LoadScene( "ld42" );
                }
                
            } break;
        }
    }
    
    void arch_update( ) {
        
        for ( int i = 0; i < arches.Length; i++ ) {
            
            Vector3 arch_position = arches[ i ].transform.position;
            
            arch_position.z -= speed * Time.deltaTime;
            
            if ( arch_position.z < cam.position.z ) {
                arch_position.z += arches.Length * arches_spacing;
            }
            
            arches[ i ].transform.position = arch_position;
        }
    }
    
    void tetromino_spawn( Vector3 position ) {
        
        for ( int i = 0; i < pieces.Length; i++ ) {
            Destroy( pieces[ i ] );
            int index = Random.Range( 0, tetrominos.Length );
            piece_indices[ i ] = index;
            pieces[ i ] = Instantiate( tetrominos[ index ], new Vector3( -15 + i * 15, 0, position.z ), Quaternion.identity );
        }
    }
    
    void tetromino_update( ) {
        
        Vector3 player_position = player.transform.position;
        float closest = 100 * 100;
        int index = -1;
        
        for ( int i = 0; i < pieces.Length; i++ ) {
            
            if ( pieces[ i ] ) {
                
                Vector3 position = pieces[ i ].transform.position;
                position.z -= speed * Time.deltaTime;
                pieces[ i ].transform.position = position;
                
                if ( !ship_piece ) {
                    Vector3 offset = position - player_position;
                    
                    if ( offset.sqrMagnitude < 10 * 10 ) {
                        
                        if ( offset.sqrMagnitude < closest ) {
                            closest = offset.sqrMagnitude;
                            index = i;
                        }
                    }
                }
            }
        }
        
        if (  index >= 0 && index < 3 ) {
            ship_piece = pieces[ index ];
            ship_piece_index = piece_indices[ index ];
            pieces[ index ] = null;
        }
    }
    
    void ship_update( ) {
        
        if ( Input.GetKeyDown( KeyCode.UpArrow ) ) {
            if ( ship_y >= -5 && ship_y < 4 ) {
                ship_y += 1;
            }
        } else if ( Input.GetKeyDown( KeyCode.DownArrow ) ) {
            if ( ship_y > -5 && ship_y <= 4 ) {
                ship_y -= 1;
            }
        } else if ( Input.GetKeyDown( KeyCode.LeftArrow ) ) {
            if ( ship_x > -5 && ship_x <= 4 ) {
                ship_x -= 1;
            }
        } else if ( Input.GetKeyDown( KeyCode.RightArrow ) ) {
            if ( ship_x >= -5 && ship_x < 4 ) {
                ship_x += 1;
            }
        }
        
        if ( Input.GetKeyDown( KeyCode.A ) || Input.GetKeyDown( KeyCode.Q ) ) {
            ship_piece_orientation++;
            ship_piece_orientation %= 4;
        } else if ( Input.GetKeyDown( KeyCode.D ) || Input.GetKeyDown( KeyCode.Space ) || Input.GetKeyDown( KeyCode.F ) ) {
            ship_piece_orientation += 4;
            ship_piece_orientation--;
            ship_piece_orientation %= 4;
        }
        
        Vector3 current_position = player.transform.position;
        Vector3 target_position = new Vector3( ship_x * 4 + 2, ship_y * 4 + 2, current_position.z );
        player.transform.position = Vector3.Lerp( current_position, target_position, 0.125f );
        
        Vector3 camera_position = cam.position;
        Vector3 camera_new = player.transform.position + camera_ship_offset;
        cam.position = Vector3.Lerp( camera_position, camera_new, 0.1f );
        camera_target.position = cam.position + camera_target_offset;
        cam.LookAt( camera_target );
    }
    
    void timer_start( Timer timer, float duration ) {
        timer.current = 0;
        timer.duration = duration;
    }
    
    void timer_update( Timer timer ) {
        timer.current += Time.deltaTime;
    }
    
    bool timer_isFinished( Timer timer, bool stopIfFinished ) {
        
        bool result = false;
        
        if ( timer.duration > 0 && timer.current >= timer.duration ) {
            result = true;
            if ( stopIfFinished ) {
                timer.duration = 0;
            }
        }
        
        return result;
    }
    
    void state_switch( State new_state ) {
        state = new_state;
        stage = Stage.init;
    }
    
    bool state_in_init_stage( ) {
        
        bool result = stage == Stage.init;
        
        if ( result ) {
            stage = Stage.running;
        }
        
        return result;
    }
    
    bool does_piece_fit_in_board( Piece piece, int[ ] board, int ship_x, int ship_y ) {
        
        bool fit = true;
        int start_x = ship_x + 5 - piece.cx;
        int start_y = ship_y + 5 - piece.cy;
        int end_x = start_x + piece.w;
        int end_y = start_y + piece.h;
        
        // Debug.Log( "(" + start_x + ", " + start_y + ") to (" + end_x + ", " + end_y + ")" );
        if ( start_x < 0 || start_y < 0 || end_x > 10 || end_y > 10 ) {
            fit = false;
        }
        
        for ( int ty = start_y, j = 0; fit && ty < end_y; ty++, j++ ) {
            for ( int tx = start_x, i = 0; fit && tx < end_x; tx++, i++ ) {
                
                int data_index = j * 4 + i;
                int board_index = ty * 10 + tx;
                
                if ( piece.data[ data_index ] != 0 && board[ board_index ] != 0 ) {
                    fit = false;
                }
            }
        }
        
        return fit;
    }
    
    void add_piece_in_board( Piece piece, int[ ] board, int ship_x, int ship_y ) {
        
        int start_x = ship_x + 5 - piece.cx;
        int start_y = ship_y + 5 - piece.cy;
        int end_x = start_x + piece.w;
        int end_y = start_y + piece.h;
        
        for ( int ty = start_y, j = 0; ty < end_y; ty++, j++ ) {
            for ( int tx = start_x, i = 0; tx < end_x; tx++, i++ ) {
                
                if ( piece.data[ j * 4 + i ] != 0 ) {
                    Debug.Assert( board[ ty * 10 + tx ] == 0 );
                    board[ ty * 10 + tx ] = 1;
                }
            }
        }
    }
}
